﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.ParticleSwarmOptimization {

  [Item("Particle Swarm Optimization", "A particle swarm optimization algorithm.")]
  [Creatable("Algorithms")]
  [StorableClass]
  public class ParticleSwarmOptimization : EngineAlgorithm {

    #region Problem Properties
    public override Type ProblemType {
      get { return typeof(ISingleObjectiveProblem); }
    }
    public new ISingleObjectiveProblem Problem {
      get { return (ISingleObjectiveProblem)base.Problem; }
      set { base.Problem = value; }
    }
    public MultiAnalyzer Analyzer {
      get { return AnalyzerParameter.Value; }
      set { AnalyzerParameter.Value = value; }
    }
    #endregion

    #region Parameter Properties
    public ValueParameter<IntValue> SeedParameter {
      get { return (ValueParameter<IntValue>)Parameters["Seed"]; }
    }
    public ValueParameter<BoolValue> SetSeedRandomlyParameter {
      get { return (ValueParameter<BoolValue>)Parameters["SetSeedRandomly"]; }
    }
    public ValueParameter<IntValue> SwarmSizeParameter {
      get { return (ValueParameter<IntValue>)Parameters["SwarmSize"]; }
    }
    public ValueParameter<IntValue> MaxIterationsParameter {
      get { return (ValueParameter<IntValue>)Parameters["MaxIterations"]; }
    }
    public ValueParameter<DoubleValue> OmegaParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["Omega"]; }
    }
    public ValueParameter<DoubleValue> Phi_PParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["Phi_P"]; }
    }
    public ValueParameter<DoubleValue> Phi_GParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["Phi_G"]; }
    }
    public ValueParameter<MultiAnalyzer> AnalyzerParameter {
      get { return (ValueParameter<MultiAnalyzer>)Parameters["Analyzer"]; }
    }
    public ValueLookupParameter<DoubleMatrix> VelocityBoundsParameter {
      get { return (ValueLookupParameter<DoubleMatrix>)Parameters["VelocityBounds"]; }
    }
    #endregion

    #region Properties
    [Storable]
    private BestAverageWorstQualityAnalyzer qualityAnalyzer;
    #endregion

    [StorableConstructor]
    protected ParticleSwarmOptimization(bool deserializing) : base(deserializing) { }
    protected ParticleSwarmOptimization(ParticleSwarmOptimization original, Cloner cloner)
      : base(original, cloner) {
      qualityAnalyzer = cloner.Clone(original.qualityAnalyzer);
    }

    public ParticleSwarmOptimization()
      : base() {
      Parameters.Add(new ValueParameter<IntValue>("Seed", "The random seed used to initialize the new pseudo random number generator.", new IntValue(0)));
      Parameters.Add(new ValueParameter<BoolValue>("SetSeedRandomly", "True if the random seed should be set to a random value, otherwise false.", new BoolValue(true)));
      Parameters.Add(new ValueParameter<IntValue>("SwarmSize", "Size of the particle swarm.", new IntValue(10)));
      Parameters.Add(new ValueParameter<IntValue>("MaxIterations", "Maximal number of iterations.", new IntValue(1000)));
      Parameters.Add(new ValueParameter<MultiAnalyzer>("Analyzer", "The operator used to analyze each generation.", new MultiAnalyzer()));
      Parameters.Add(new ValueParameter<DoubleValue>("Omega", "Weight for particle's velocity vector.", new DoubleValue(-0.2)));
      Parameters.Add(new ValueParameter<DoubleValue>("Phi_P", "Weight for particle's personal best position.", new DoubleValue(-0.01)));
      Parameters.Add(new ValueParameter<DoubleValue>("Phi_G", "Weight for global best position.", new DoubleValue(3.7)));
      Parameters.Add(new ValueLookupParameter<DoubleMatrix>("VelocityBounds", "Maximum Velocity in every dimension", new DoubleMatrix(new double[,] { { -1, 1 } })));

      RandomCreator randomCreator = new RandomCreator();
      VariableCreator variableCreator = new VariableCreator();
      SolutionsCreator solutionsCreator = new SolutionsCreator();
      CombinedOperator particleCreator = new CombinedOperator();
      Placeholder evaluatorPlaceholder = new Placeholder();
      Assigner bestPersonalQualityAssigner = new Assigner();
      BestPointInitializer bestPositionInitializer = new BestPointInitializer();
      Placeholder analyzerPlaceholder = new Placeholder();
      UniformSubScopesProcessor uniformSubScopeProcessor = new UniformSubScopesProcessor();
      ParticleUpdater particleUpdater = new ParticleUpdater();
      Placeholder evaluatorPlaceholder2 = new Placeholder();
      SwarmUpdater swarmUpdater = new SwarmUpdater();
      Placeholder analyzerPlaceholder2 = new Placeholder();
      IntCounter currentIterationCounter = new IntCounter();
      Comparator currentIterationComparator = new Comparator();
      ConditionalBranch conditionalBranch = new ConditionalBranch();

      OperatorGraph.InitialOperator = randomCreator;

      randomCreator.SetSeedRandomlyParameter.Value = null;
      randomCreator.SeedParameter.Value = null;
      randomCreator.Successor = variableCreator;

      variableCreator.CollectedValues.Add(new ValueParameter<IntValue>("CurrentIteration", new IntValue(0)));
      variableCreator.Successor = solutionsCreator;

      solutionsCreator.NumberOfSolutionsParameter.ActualName = "SwarmSize";
      solutionsCreator.EvaluatorParameter.Value = evaluatorPlaceholder;
      solutionsCreator.SolutionCreatorParameter.Value = particleCreator;
      solutionsCreator.Successor = bestPositionInitializer;

      InitializeParticleCreator(particleCreator);

      evaluatorPlaceholder.Name = "(Evaluator)";
      evaluatorPlaceholder.OperatorParameter.ActualName = "Evaluator";
      evaluatorPlaceholder.Successor = bestPersonalQualityAssigner;

      bestPersonalQualityAssigner.LeftSideParameter.ActualName = "PersonalBestQuality";
      bestPersonalQualityAssigner.RightSideParameter.ActualName = "Quality";

      bestPositionInitializer.Successor = analyzerPlaceholder;

      analyzerPlaceholder.Name = "(Analyzer)";
      analyzerPlaceholder.OperatorParameter.ActualName = "Analyzer";
      analyzerPlaceholder.Successor = uniformSubScopeProcessor;

      uniformSubScopeProcessor.Operator = particleUpdater;
      uniformSubScopeProcessor.Successor = analyzerPlaceholder2;

      particleUpdater.Successor = evaluatorPlaceholder2;

      evaluatorPlaceholder2.Name = "(Evaluator)";
      evaluatorPlaceholder2.OperatorParameter.ActualName = "Evaluator";
      evaluatorPlaceholder2.Successor = swarmUpdater;

      analyzerPlaceholder2.Name = "(Analyzer)";
      analyzerPlaceholder2.OperatorParameter.ActualName = "Analyzer";
      analyzerPlaceholder2.Successor = currentIterationCounter;

      currentIterationCounter.Name = "CurrentIteration++";
      currentIterationCounter.ValueParameter.ActualName = "CurrentIteration";
      currentIterationCounter.Successor = currentIterationComparator;

      currentIterationComparator.LeftSideParameter.ActualName = "CurrentIteration";
      currentIterationComparator.Comparison = new Comparison(ComparisonType.Less);
      currentIterationComparator.RightSideParameter.ActualName = "MaxIterations";
      currentIterationComparator.ResultParameter.ActualName = "ContinueIteration";
      currentIterationComparator.Successor = conditionalBranch;

      conditionalBranch.Name = "ContinueIteration?";
      conditionalBranch.ConditionParameter.ActualName = "ContinueIteration";
      conditionalBranch.TrueBranch = uniformSubScopeProcessor;

      InitializeAnalyzers();
      UpdateAnalyzers();
    }

    private static void InitializeParticleCreator(CombinedOperator particleCreator) {
      Placeholder positionCreator = new Placeholder();
      Assigner personalBestPositionAssigner = new Assigner();
      UniformRandomRealVectorCreator velocityCreator = new UniformRandomRealVectorCreator();

      particleCreator.Name = "Particle Creator";
      particleCreator.OperatorGraph.InitialOperator = positionCreator;

      positionCreator.Name = "(SolutionCreator)";
      positionCreator.OperatorParameter.ActualName = "SolutionCreator";
      positionCreator.Successor = personalBestPositionAssigner;

      personalBestPositionAssigner.LeftSideParameter.ActualName = "PersonalBestPoint";
      personalBestPositionAssigner.RightSideParameter.ActualName = "Point";
      personalBestPositionAssigner.Successor = velocityCreator;

      velocityCreator.LengthParameter.ActualName = "ProblemSize";
      velocityCreator.BoundsParameter.ActualName = "VelocityBounds";
      velocityCreator.RealVectorParameter.ActualName = "Velocity";
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ParticleSwarmOptimization(this, cloner);
    }

    public override void Prepare() {
      if (Problem != null) base.Prepare();
    }

    #region Events
    protected override void OnProblemChanged() {
      UpdateAnalyzers();
      ParameterizeAnalyzers();
      Problem.Evaluator.QualityParameter.ActualNameChanged += new EventHandler(Evaluator_QualityParameter_ActualNameChanged);
      base.OnProblemChanged();
    }

    #endregion

    #region Helpers
    private void Evaluator_QualityParameter_ActualNameChanged(object sender, EventArgs e) {
    }

    private void InitializeAnalyzers() {
      qualityAnalyzer = new BestAverageWorstQualityAnalyzer();
      ParameterizeAnalyzers();
    }

    private void ParameterizeAnalyzers() {
      if (Problem != null) {
        qualityAnalyzer.MaximizationParameter.ActualName = Problem.MaximizationParameter.Name;
        qualityAnalyzer.QualityParameter.ActualName = Problem.Evaluator.QualityParameter.ActualName;
        qualityAnalyzer.BestKnownQualityParameter.ActualName = Problem.BestKnownQualityParameter.Name;
      }
    }

    private void UpdateAnalyzers() {
      Analyzer.Operators.Clear();
      if (Problem != null) {
        foreach (IAnalyzer analyzer in Problem.Operators.OfType<IAnalyzer>())
          Analyzer.Operators.Add(analyzer);
      }
      Analyzer.Operators.Add(qualityAnalyzer);
    }

    #endregion
  }
}
